@@ -107,16 +107,21 @@ fn main() {
             src_dir: "vendor/tree-sitter-elixir-src",
             extra_files: vec!["scanner.cc"],
         },
-        TreeSitterParser {
-            name: "tree-sitter-gleam",
-            src_dir: "vendor/tree-sitter-gleam-src",
-            extra_files: vec![],
-        },
         TreeSitterParser {
             name: "tree-sitter-elm",
             src_dir: "vendor/tree-sitter-elm-src",
             extra_files: vec!["scanner.cc"],
         },
+        TreeSitterParser {
+            name: "tree-sitter-elvish",
+            src_dir: "vendor/tree-sitter-elvish-src",
+            extra_files: vec![],
+        },
+        TreeSitterParser {
+            name: "tree-sitter-gleam",
+            src_dir: "vendor/tree-sitter-gleam-src",
+            extra_files: vec![],
+        },
         TreeSitterParser {
             name: "tree-sitter-go",
             src_dir: "vendor/tree-sitter-go-src",
