@@ -14,35 +14,28 @@
 #![allow(clippy::too_many_arguments)]
 
 mod constants;
-mod context;
-mod dijkstra;
+mod diff;
+mod display;
 mod files;
-mod graph;
-mod guess_language;
-mod hunks;
-mod inline;
 mod line_parser;
 mod lines;
-mod myers_diff;
 mod options;
+mod parse;
 mod positions;
-mod side_by_side;
-mod sliders;
-mod style;
 mod summary;
-mod syntax;
-mod tree_sitter_parser;
-mod unchanged;
 
 #[macro_use]
 extern crate log;
 
-use crate::hunks::{matched_pos_to_hunks, merge_adjacent};
-use context::opposite_positions;
-use files::read_files_or_die;
-use guess_language::guess;
+use crate::diff::{dijkstra, unchanged};
+use crate::display::hunks::{matched_pos_to_hunks, merge_adjacent};
+use crate::parse::syntax;
+use diff::changes::ChangeMap;
+use display::context::opposite_positions;
+use files::{is_probably_binary, read_files_or_die, read_or_die, relative_paths_in_either};
 use log::info;
 use mimalloc::MiMalloc;
+use parse::guess_language::guess;
 
 /// The global allocator used by difftastic.
 ///
@@ -51,21 +44,17 @@ use mimalloc::MiMalloc;
 #[global_allocator]
 static GLOBAL: MiMalloc = MiMalloc;
 
-use options::{should_use_color, DisplayMode, Mode};
-use sliders::fix_all_sliders;
+use diff::sliders::fix_all_sliders;
+use options::{DisplayMode, DisplayOptions, Mode};
+use rayon::prelude::*;
 use std::{env, path::Path};
-use style::BackgroundColor;
 use summary::{DiffResult, FileContent};
 use syntax::init_next_prev;
 use typed_arena::Arena;
-use walkdir::WalkDir;
 
 use crate::{
-    dijkstra::mark_syntax,
-    files::{is_probably_binary, read_or_die},
-    lines::MaxLine,
-    syntax::init_all_info,
-    tree_sitter_parser as tsp,
+    dijkstra::mark_syntax, lines::MaxLine, parse::syntax::init_all_info,
+    parse::tree_sitter_parser as tsp,
 };
 
 extern crate pretty_env_logger;
@@ -134,20 +123,14 @@ fn main() {
         Mode::Diff {
             node_limit,
             byte_limit,
-            print_unchanged,
+            display_options,
             missing_as_empty,
-            display_mode,
-            background_color,
-            color_output,
-            display_width,
-            display_path,
             language_override,
             lhs_path,
             rhs_path,
-            ..
+            lhs_display_path,
+            rhs_display_path,
         } => {
-            let use_color = should_use_color(color_output);
-
             let lhs_path = Path::new(&lhs_path);
             let rhs_path = Path::new(&rhs_path);
 
@@ -163,41 +146,30 @@ fn main() {
             }
 
             if lhs_path.is_dir() && rhs_path.is_dir() {
-                for diff_result in diff_directories(
+                diff_directories(
                     lhs_path,
                     rhs_path,
-                    missing_as_empty,
+                    &display_options,
                     node_limit,
                     byte_limit,
                     language_override,
-                ) {
-                    print_diff_result(
-                        display_width,
-                        use_color,
-                        display_mode,
-                        background_color,
-                        print_unchanged,
-                        &diff_result,
-                    );
-                }
+                )
+                .for_each(|diff_result| {
+                    print_diff_result(&display_options, &diff_result);
+                });
             } else {
                 let diff_result = diff_file(
-                    &display_path,
+                    &lhs_display_path,
+                    &rhs_display_path,
                     lhs_path,
                     rhs_path,
+                    &display_options,
                     missing_as_empty,
                     node_limit,
                     byte_limit,
                     language_override,
                 );
-                print_diff_result(
-                    display_width,
-                    use_color,
-                    display_mode,
-                    background_color,
-                    print_unchanged,
-                    &diff_result,
-                );
+                print_diff_result(&display_options, &diff_result);
             }
         }
     };
@@ -205,19 +177,23 @@ fn main() {
 
 /// Print a diff between two files.
 fn diff_file(
-    display_path: &str,
+    lhs_display_path: &str,
+    rhs_display_path: &str,
     lhs_path: &Path,
     rhs_path: &Path,
+    display_options: &DisplayOptions,
     missing_as_empty: bool,
     node_limit: u32,
     byte_limit: usize,
-    language_override: Option<guess_language::Language>,
+    language_override: Option<parse::guess_language::Language>,
 ) -> DiffResult {
     let (lhs_bytes, rhs_bytes) = read_files_or_die(lhs_path, rhs_path, missing_as_empty);
     diff_file_content(
-        display_path,
+        lhs_display_path,
+        rhs_display_path,
         &lhs_bytes,
         &rhs_bytes,
+        display_options.tab_width,
         node_limit,
         byte_limit,
         language_override,
@@ -225,16 +201,19 @@ fn diff_file(
 }
 
 fn diff_file_content(
-    display_path: &str,
+    lhs_display_path: &str,
+    rhs_display_path: &str,
     lhs_bytes: &[u8],
     rhs_bytes: &[u8],
+    tab_width: usize,
     node_limit: u32,
     byte_limit: usize,
-    language_override: Option<guess_language::Language>,
+    language_override: Option<parse::guess_language::Language>,
 ) -> DiffResult {
     if is_probably_binary(lhs_bytes) || is_probably_binary(rhs_bytes) {
         return DiffResult {
-            path: display_path.into(),
+            lhs_display_path: lhs_display_path.into(),
+            rhs_display_path: rhs_display_path.into(),
             language: None,
             lhs_src: FileContent::Binary(lhs_bytes.to_vec()),
             rhs_src: FileContent::Binary(rhs_bytes.to_vec()),
@@ -244,12 +223,13 @@ fn diff_file_content(
     }
 
     // TODO: don't replace tab characters inside string literals.
+    let tab_as_spaces = " ".repeat(tab_width);
     let mut lhs_src = String::from_utf8_lossy(lhs_bytes)
         .to_string()
-        .replace('\t', "        ");
+        .replace('\t', &tab_as_spaces);
     let mut rhs_src = String::from_utf8_lossy(rhs_bytes)
         .to_string()
-        .replace('\t', "        ");
+        .replace('\t', &tab_as_spaces);
 
     // Ignore the trailing newline, if present.
     // TODO: highlight if this has changes (#144).
@@ -262,7 +242,7 @@ fn diff_file_content(
     }
 
     // TODO: take a Path directly instead.
-    let path = Path::new(&display_path);
+    let guess_path = Path::new(&rhs_display_path);
 
     // Take the larger of the two files when guessing the
     // language. This is useful when we've added or removed a whole
@@ -272,14 +252,15 @@ fn diff_file_content(
     } else {
         &rhs_src
     };
-    let language = language_override.or_else(|| guess(path, guess_src));
+    let language = language_override.or_else(|| guess(guess_path, guess_src));
     let lang_config = language.map(tsp::from_language);
 
     if lhs_bytes == rhs_bytes {
         // If the two files are completely identical, return early
         // rather than doing any more work.
         return DiffResult {
-            path: display_path.into(),
+            lhs_display_path: lhs_display_path.into(),
+            rhs_display_path: rhs_display_path.into(),
             language: lang_config.map(|l| l.name.into()),
             lhs_src: FileContent::Text("".into()),
             rhs_src: FileContent::Text("".into()),
@@ -305,10 +286,11 @@ fn diff_file_content(
 
             init_all_info(&lhs, &rhs);
 
+            let mut change_map = ChangeMap::default();
             let possibly_changed = if env::var("DFT_DBG_KEEP_UNCHANGED").is_ok() {
                 vec![(lhs.clone(), rhs.clone())]
             } else {
-                unchanged::mark_unchanged(&lhs, &rhs)
+                unchanged::mark_unchanged(&lhs, &rhs, &mut change_map)
             };
 
             let possibly_changed_max = max_num_nodes(&possibly_changed);
@@ -333,15 +315,16 @@ fn diff_file_content(
                     mark_syntax(
                         lhs_section_nodes.get(0).copied(),
                         rhs_section_nodes.get(0).copied(),
+                        &mut change_map,
                     );
 
                     let language = language.unwrap();
-                    fix_all_sliders(language, &lhs_section_nodes);
-                    fix_all_sliders(language, &rhs_section_nodes);
+                    fix_all_sliders(language, &lhs_section_nodes, &mut change_map);
+                    fix_all_sliders(language, &rhs_section_nodes, &mut change_map);
                 }
 
-                let lhs_positions = syntax::change_positions(&lhs);
-                let rhs_positions = syntax::change_positions(&rhs);
+                let lhs_positions = syntax::change_positions(&lhs, &change_map);
+                let rhs_positions = syntax::change_positions(&rhs, &change_map);
                 (Some(ts_lang.name.into()), lhs_positions, rhs_positions)
             }
         }
@@ -353,7 +336,8 @@ fn diff_file_content(
     };
 
     DiffResult {
-        path: display_path.into(),
+        lhs_display_path: lhs_display_path.into(),
+        rhs_display_path: rhs_display_path.into(),
         language: lang_name,
         lhs_src: FileContent::Text(lhs_src),
         rhs_src: FileContent::Text(rhs_src),
@@ -371,43 +355,39 @@ fn diff_file_content(
 fn diff_directories<'a>(
     lhs_dir: &'a Path,
     rhs_dir: &'a Path,
-    missing_as_empty: bool,
+    display_options: &DisplayOptions,
     node_limit: u32,
     byte_limit: usize,
-    language_override: Option<guess_language::Language>,
-) -> impl Iterator<Item = DiffResult> + 'a {
-    WalkDir::new(lhs_dir)
-        .into_iter()
-        .filter_map(Result::ok)
-        .map(|entry| entry.into_path())
-        .filter(|lhs_path| !lhs_path.is_dir())
-        .map(move |lhs_path| {
-            info!("LHS path is {:?} inside {:?}", lhs_path, lhs_dir);
+    language_override: Option<parse::guess_language::Language>,
+) -> impl ParallelIterator<Item = DiffResult> + 'a {
+    let display_options = display_options.clone();
 
-            let rel_path = lhs_path.strip_prefix(lhs_dir).unwrap();
-            let rhs_path = Path::new(rhs_dir).join(rel_path);
+    // We greedily list all files in the directory, and then diff them
+    // in parallel. This is assuming that diffing is slower than
+    // enumerating files, so it benefits more from parallelism.
+    let paths = relative_paths_in_either(lhs_dir, rhs_dir);
 
-            diff_file(
-                &rel_path.to_string_lossy(),
-                &lhs_path,
-                &rhs_path,
-                missing_as_empty,
-                node_limit,
-                byte_limit,
-                language_override,
-            )
-        })
+    paths.into_par_iter().map(move |rel_path| {
+        info!("Relative path is {:?} inside {:?}", rel_path, lhs_dir);
+
+        let lhs_path = Path::new(lhs_dir).join(&rel_path);
+        let rhs_path = Path::new(rhs_dir).join(&rel_path);
+
+        diff_file(
+            &rel_path.to_string_lossy(),
+            &rel_path.to_string_lossy(), // todo
+            &lhs_path,
+            &rhs_path,
+            &display_options,
+            true,
+            node_limit,
+            byte_limit,
+            language_override,
+        )
+    })
 }
 
-// TODO: factor out a DiffOptions struct.
-fn print_diff_result(
-    display_width: usize,
-    use_color: bool,
-    display_mode: DisplayMode,
-    background: BackgroundColor,
-    print_unchanged: bool,
-    summary: &DiffResult,
-) {
+fn print_diff_result(display_options: &DisplayOptions, summary: &DiffResult) {
     match (&summary.lhs_src, &summary.rhs_src) {
         (FileContent::Text(lhs_src), FileContent::Text(rhs_src)) => {
             let opposite_to_lhs = opposite_positions(&summary.lhs_positions);
@@ -424,10 +404,17 @@ fn print_diff_result(
 
             let lang_name = summary.language.clone().unwrap_or_else(|| "Text".into());
             if hunks.is_empty() {
-                if print_unchanged {
+                if display_options.print_unchanged {
                     println!(
                         "{}",
-                        style::header(&summary.path, 1, 1, &lang_name, use_color, background)
+                        display::style::header(
+                            &summary.lhs_display_path,
+                            &summary.rhs_display_path,
+                            1,
+                            1,
+                            &lang_name,
+                            display_options
+                        )
                     );
                     if lang_name == "Text" || summary.lhs_src == summary.rhs_src {
                         // TODO: there are other Text names now, so
@@ -440,28 +427,26 @@ fn print_diff_result(
                 return;
             }
 
-            match display_mode {
+            match display_options.display_mode {
                 DisplayMode::Inline => {
-                    inline::print(
+                    display::inline::print(
                         lhs_src,
                         rhs_src,
+                        display_options,
                         &summary.lhs_positions,
                         &summary.rhs_positions,
                         &hunks,
-                        &summary.path,
+                        &summary.lhs_display_path,
+                        &summary.rhs_display_path,
                         &lang_name,
-                        use_color,
-                        background,
                     );
                 }
                 DisplayMode::SideBySide | DisplayMode::SideBySideShowBoth => {
-                    side_by_side::print(
+                    display::side_by_side::print(
                         &hunks,
-                        display_width,
-                        use_color,
-                        display_mode,
-                        background,
-                        &summary.path,
+                        display_options,
+                        &summary.lhs_display_path,
+                        &summary.rhs_display_path,
                         &lang_name,
                         lhs_src,
                         rhs_src,
@@ -473,10 +458,17 @@ fn print_diff_result(
         }
         (FileContent::Binary(lhs_bytes), FileContent::Binary(rhs_bytes)) => {
             let changed = lhs_bytes != rhs_bytes;
-            if print_unchanged || changed {
+            if display_options.print_unchanged || changed {
                 println!(
                     "{}",
-                    style::header(&summary.path, 1, 1, "binary", use_color, background)
+                    display::style::header(
+                        &summary.lhs_display_path,
+                        &summary.rhs_display_path,
+                        1,
+                        1,
+                        "binary",
+                        display_options
+                    )
                 );
                 if changed {
                     println!("Binary contents changed.");
@@ -489,7 +481,14 @@ fn print_diff_result(
             // We're diffing a binary file against a text file.
             println!(
                 "{}",
-                style::header(&summary.path, 1, 1, "binary", use_color, background)
+                display::style::header(
+                    &summary.lhs_display_path,
+                    &summary.rhs_display_path,
+                    1,
+                    1,
+                    "binary",
+                    display_options
+                )
             );
             println!("Binary contents changed.");
         }
@@ -522,15 +521,17 @@ fn max_num_nodes(roots_vec: &[(Vec<&syntax::Syntax>, Vec<&syntax::Syntax>)]) ->
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::options::{DEFAULT_BYTE_LIMIT, DEFAULT_NODE_LIMIT};
+    use crate::options::{DEFAULT_BYTE_LIMIT, DEFAULT_NODE_LIMIT, DEFAULT_TAB_WIDTH};
 
     #[test]
     fn test_diff_identical_content() {
         let s = "foo";
         let res = diff_file_content(
+            "foo.el",
             "foo.el",
             s.as_bytes(),
             s.as_bytes(),
+            DEFAULT_TAB_WIDTH,
             DEFAULT_NODE_LIMIT,
             DEFAULT_BYTE_LIMIT,
             None,
